/*****************************************************************************
	
	Batterieinnenwiderstandsmessgert 

	mit "Controller ATMega328_Mini" 
	display EA_DOGM162, Encoder und vier Tasten, 8 MHz extern

	Funktions: Es wird der Innenwiderstand eines Akkus oder Batterie gemessen.
	Spannungen bis 4,5 V sind mglich.
	
	Es wird die Spannung U_0 im Leerlauf gemessen und die Spannung U_1 im Anschluss
	fr eine Dauer von ca 300 ms. Fr 700 ms wir der Akku nicht belastet. 
	Der Laststrom kann mit der Key3 voreingestellt werden. Der Default Wert ist 2000 mA.
	
	Der tatschlich flieende Strom wird ber einen Shuntwiderstand von 0,1 Ohm gemessen. 
	Der Controller steuer ber einen 12 Bit D/A Wander die Spannung am Gate eines HEXFET, 
	der den Laststom voreinstellt.

	taster1:	die Messung im Pulsbetrieb starten, Ri wird in der zweiten Zeile angezeigt.
	taster2:	Stopp der Messung, Laststrom wird ausgeschaltet
	taster3:	Laststrom voreinstellen, mit SET des Drehencoders besttigen!
	taster4:	-
	encoder:	Laststrom erhhen/reduzieren ( 100mA bis 10A)
	SET:		quittieren

	License:
    Copyright (C) 2016 Clemens Verstappen DL3ETW

	This program is free software; you can redistribute it and/or modify it under the terms  
 	of the GNU General Public License as published by the Free Software Foundation; either 
	version 3 of the License, or (at your option) any later version.	 
	This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
  	without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU General Public License for more details.You should have received a copy of 
	the GNU General Public License along with this program; if not, see <http://www.gnu.org/licenses/>. 

	Code for i2c and uart from Peter Fleury.
	http://homepage.hispeed.ch/peterfleury/avr-software.html
	
	Code supporting the EADOGM LCD from Peter Dannegger, from www.mikrocontroller.net

	history:
	V1.00 - first version Januar 2016
	V1.00 - Publication, April 2016 
*****************************************************************************/
#include 	"lcd_drv.h"						//  LCD EADOG config
#include 	"LCD_DRV.C"
#include 	"i2cmaster.h"
#include 	<inttypes.h>                	// general lib
#include 	<avr/io.h>
#include 	<avr/interrupt.h>
#include 	<stdlib.h>						
#include 	<util/delay.h>	
#include 	"uart.h"
			
#define 	UART_BAUD_RATE      9600      
#define 	F_CPU 				8000000UL 	// xtal frequency

#define		LEBENSLICHT 		PD2			// port pin for LED
#define		TESTPIN 			PC3			// 

#define 	KEY1				0b00000001	// mask
#define 	KEY2				0b00000010	// mask
#define 	KEY3				0b00001000	// mask
#define 	KEY4				0b00010000	// mask
#define 	KEY5				0b00100000	// mask	encoder switch
#define 	MCP4226				0b11000000	// A2 device address of DAC Converter, see datasheet
#define		TIMER_100MS			0x030E		// 0x030E time value for 100ms interrups of realtime_update 

volatile 	uint8_t 			display_timer = 20;		// start wert fr den Anfangswert 
volatile 	uint16_t 			load_counter;			// Zhler fr die Iterationsstufen
volatile 	uint16_t 			adc[2+1];				// Sepatates Array fr die Batterie auf dem Controller328_Mini
volatile 	uint16_t 			ad_spannung_strom[4+1];	// 1: U_0, 2: I_0, 3: U_1, 4: I_1 
volatile 	uint16_t 			dac_value, last_strom = 2000;		// Anfangswerrt der Last = 2000 mA

volatile	float				U_0, U_1, I_0, I_1, R_i; 	// Leerlaufspannung, Lastspannung, Laststrom und Innenwiderstand Ri
volatile 	float 				adc_result, adc_current, adc_voltage;
char		ascii[7+1][16]; 						// ascii[0] Spannung Controller Batterie, 
													// ascii[1] currend, 
													// ascii[2] voltage, ascii[3] R_I,
													 
volatile struct BitVariablen						// status of keys 
{									
	int 	key1					:1;				// key1
	int 	key2					:1;
	int 	key3					:1;
	int 	key4					:1;
	int		key5					:1;
	int 	encoder_ratation		:1;				// encoder flag, if turned
	int		realtime_update 		:1;				// flag for main routine
	int		display_timer			:1;
	int		measure			 		:1;
	int		display			 		:1;
}status = {1};

void lcd_message(uint16_t disp_value); 				// this routine calculates the telegram
void read_ad_channel();								// 
void set_load_current(uint16_t tmp);				// load via I2C DAC 
void dac_power_down();								// load via I2C DAC 
void display_voltage_current(uint8_t voltage, uint8_t current); 	
uint32_t calc_dac_value(uint32_t testo);			// calculate DAC value from mAh of accu

// *************************************************************************

int main (void) 
{
	DDRC 	|= (1<<TESTPIN);							// Testpin output = ADC3 Achtung
	DDRD 	|= (1<<LEBENSLICHT);	

	PORTB 	|= ((1<<PB0) | (1 << PB1) | (1 << PB3) | (1 << PB4) | (1 << PB5));	// Pull Ups einschalten

	OCR1A	 = TIMER_100MS;										// timer 1, compare match
	TCCR1B	|= ((1 << WGM12) | (1 << CS12) | (1 << CS10)); 		// clear on compare match, divide by 1024
	TIMSK1	|= (1 << OCIE1A);									// timer1 compare match 1 interrupt enable

	PCMSK0 	|= (1 << PCINT0) |(1 << PCINT1) |(1 << PCINT3) |(1 << PCINT4) |(1 << PCINT5);	// PIN change interrup 
	PCICR	|= (1 << PCIE0);							// PIN change interrrupt enable

	EICRA	|= (1<<ISC11)|(1<<ISC10);					// steigende Flanke vom externen interrupt fr den Encoder
	EIMSK	|= (1<<INT1);								// enable external interrupt1	

	ADMUX	|= (1 << REFS1) | (1 << REFS0);				// Internal 1.1V Referenz, 
	ADCSRA 	|= (1 << ADEN); 							// Vorteiler auf 2

	i2c_init();                                			// init I2C interface
	lcd_init();											// init LCD
	uart_init( UART_BAUD_SELECT(UART_BAUD_RATE,F_CPU) );

// ------------------------------------------------ 

	sei();										// enable interrups
	status.display_timer = 1;
	dac_power_down();

// ------------------------------------------------ 

	lcd_clear();								// welcome message on LCD
	lcd_pos( 0, 0 );
	lcd_text( "Innenwiderstand-" );
	lcd_pos( 1, 0 );
  	lcd_text( "   Messgeraet   " );
	_delay_ms(1000);	
	
// ------------------------------------------------

	read_ad_channel();							// Nur einmal die Spannung des Controller328_Mini Akku auslesen 
	adc_result = adc[0] * 4.73 / 1024;			// Mit Korrekturwert versehen.
	dtostrf (adc_result, 2,2, ascii[0]);		// Ein String bitte, zwei digits links des Kommas und zwei rechts
	lcd_pos( 1,0);
  	lcd_text( "U-batt:         " );
	lcd_pos( 1,8);
	lcd_text((char*) ascii[0]);
	lcd_pos( 1,13);
  	lcd_text("V");
	_delay_ms(1000);
													
// ------------------------------------------------ 
	status.key1 = 0;
  	while(1) 											// Endlosschleife
	{
		if (status.realtime_update)						// Interrupt fr die Hauptschleife jede 100ms
		{
			PORTD ^= (1<<LEBENSLICHT);					// toggle lebenslicht
			if (status.measure)							// wenn Messung gestartet wurde
			{
				load_counter++;							// 100ms pro countervalue
				if (load_counter == 1)					// starte 
				{
					set_load_current(0);				// kein Laststrom
					read_ad_channel();					// lese Werte fr Leerlaufspannung U_0
					ad_spannung_strom[1]  = adc[2];			// 1: U_0, 2: I_0, Leerlaufwerte fr spter merken 
					ad_spannung_strom[2]  = adc[1];			
					dac_value = calc_dac_value(last_strom);	// rechne Entladestrom aus
					set_load_current(dac_value);			// schalte nun Entladestrom ein
				}

				if (load_counter == 4)					// starte 300 ms
				{
					read_ad_channel();					// lese erneut alle Werte nun unter Last
					set_load_current(0);				// schalte Strompuls wieder aus
					ad_spannung_strom[3]  = adc[2];		// 3: U_1, 4: I_1 
					ad_spannung_strom[4]  = adc[1];		// sichere Werte fr spter
				}

				if (load_counter == 9)					// nun ist eine Sekunde um
														// ein wening rechnen und Ergebnisn anzeigen
				{
					display_voltage_current(2,1); 	

					U_0 = ad_spannung_strom[1] * 4.4 / 1024;	// Rechnung dauert sehr lange typ. 1ms
					I_0	= ad_spannung_strom[2] * 11.0 / 1024;
					U_1 = ad_spannung_strom[3] * 4.4 / 1024;
					I_1 = ad_spannung_strom[4] * 11.0 / 1024;

					I_0 = 0;							// wird nicht verwendet
					R_i = ((U_0 - U_1) / (I_1 - I_0)); 

					dtostrf ((R_i), 1,3,(char*) ascii[3]);	// one digits left of decimal, three right
					lcd_pos( 1,0);							
					lcd_text("Ri=            " );		// zeige Innenwiderstand an
					lcd_pos( 1,3);
					lcd_text((char*) ascii[3]);
					lcd_pos( 1,9);
					lcd_text("Ohm    ");

					load_counter = 0;					// nun wieder von vorne eine Sekunde ist um
				}
			}
			else										// falls noch nicht gemessen wird, zeige nur U_0 und I_0 regelmgig an
			{
				load_counter++;
				if (load_counter == 5)					// alle 500 ms anzeigen
				{
					read_ad_channel();					// lese alle Werte
					display_voltage_current(2,1); 	
					load_counter = 0;
				}
			}

// -------------------------------------------------
			if (status.key1)							// TASTER1	START
			{
				lcd_pos( 1, 0 );
 				lcd_text("  START         ");			// commend 2. LCD line
 
				lcd_pos( 1, 0 );
 				lcd_text("Start Messung   ");			// commend 2. LCD line
				_delay_ms(1000);
				
				status.measure = 1;
				load_counter = 0;
				
				status.display_timer = 1;				// Menu
				status.key1 = 0;
			}

// -------------------------------------------------
			if (status.key2)							// TASTER2  STOP
			{
				lcd_pos( 1, 0 );
 				lcd_text("    STOP        ");			// commend 2. LCD line
 
				status.measure = 0;
				dac_power_down();
				load_counter = 0;

 				status.display_timer = 1;
				status.key2 = 0;
			}

// -------------------------------------------------
			if (status.key3)							// TASTER3 C
			{
				lcd_pos( 0, 0 );
  				lcd_text( "Laststrom eingeb" );

				itoa(last_strom,(char*) ascii[4],10);
				lcd_pos( 1, 0 );
				lcd_text( "I/mA =          ");
				lcd_pos( 1,8);
				lcd_text((char*) ascii[4]);
	
				status.display_timer = 1;
//				status.key3 = 0;
			}

// -------------------------------------------------			
			if (status.key4)							// TASTER4  not used
			{
				lcd_pos( 1, 0 );
 				lcd_text("            key4");			

				status.display_timer = 1;
				status.key4 = 0;
			}
// -------------------------------------------------			
			if (status.key5)							// TASTER5
			{
				lcd_pos( 1, 0 );
 				lcd_text("o.k. stored     ");			// commend 2. LCD line

				status.display_timer = 1;
				status.key5 = 0;
				status.key3 = 0;
			}
// -------------------------------------------------	// Encoder gedreht

			if (status.encoder_ratation == 1)			// encoder UP
			{
				status.encoder_ratation = 0;
			}
// -------------------------------------------------

			if (status.display_timer)					// overwrite the  2. LCD line
			{											// with menu for four switches
				display_timer--;							
				if (display_timer == 0)
				{
					lcd_pos( 1, 0 );
 					lcd_text("Start Stop I  --");
					display_timer = 20;
					status.display_timer = 0;
				}
			}
		status.realtime_update = 0;
 	 	} // interrup close
	} // main while close
} // main close


void display_voltage_current(uint8_t voltage, uint8_t current) 	// hier die erste Zeile des LCD anzeigen 
{		
float	f_voltage, f_current;

	f_voltage = adc[voltage] * 4.4 / 1024;		// die Spannung nun als float,  4,35
	f_current = adc[current] * 11.0 / 1024;		// den Strom nun als float 11,00 

	dtostrf (f_voltage, 1,2,(char*) ascii[2]);	// eine Stelle links, zwei rechts vom Komma 
	lcd_pos( 0,0);								
	lcd_text("U=              " );
	lcd_pos( 0,2);
	lcd_text((char*) ascii[2]);
	lcd_pos( 0,6);
	lcd_text("V");

	dtostrf (f_current, 1,2,(char*) ascii[1]);	// eine Stelle links, zwei rechts vom Komma
	lcd_pos( 0,9);
	lcd_text("I=      " );
	lcd_pos( 0,11);
	lcd_text((char*) ascii[1]);
	lcd_pos( 0,15);
	lcd_text("A");
	
}			

void read_ad_channel(void)
{
	uint8_t	i,k;
	uint16_t result=0;
	for(k=0;k<3;k++)						// convertiere ADC0, ADC1 und ADC2
	{
		ADMUX = 0b11000000 + k;				// select ADC
		result = 0;
		for(i=0;i<50;i++)					// Mittelung von 50 
		{
			ADCSRA |= (1<<ADSC);			// starte Wandlung
			while(ADCSRA & (1<<ADSC));		// polling
			result += ADCW;					// Summe bilden
		}	
		adc[k] = result/50 ;				// teilen und ins Array kopieren
	}
}

void lcd_message(uint16_t disp_value) 		// diese Routine berechnet das telegram fr das LC 
{		
	ultoa (disp_value, (char*) ascii, 10); 
	lcd_pos( 0, 0 );
  	lcd_text( "Encoder         " );
	lcd_pos( 0, 9);
	lcd_text((char*) ascii);
}			

void set_load_current(uint16_t tmp)			// stelle den Laststrom ein 
{
uint8_t dac_l, dac_h;
	
	dac_l = tmp;
	dac_h = tmp >> 8;

	i2c_start_wait(MCP4226+I2C_WRITE);   	
	i2c_write(dac_h);						// high byte 
	i2c_write(dac_l);						// low byte 
	i2c_stop();
}

void dac_power_down()					// schalte DAC aus 
{
uint8_t dac_l, dac_h;
	
	dac_l = 0x00;
	dac_h = 0x00;
	dac_h |= 0b00010000 ;				// dieses Bit PD0 im Register setzen zum Ausschalten

	i2c_start_wait(MCP4226+I2C_WRITE);  
	i2c_write(dac_h);					// high byte 
	i2c_write(dac_l);					// low byte 
	i2c_stop();
}

uint32_t calc_dac_value(uint32_t testo)	// berechne aus Kapazitt des Akkus den Laststrom
{
	uint32_t tmp;
	tmp = (testo * 248 / 1000); 
	return (tmp);
}

SIGNAL(PCINT0_vect)						// interrupt routine fr pin change
{										// jeder Tastendruck lst einen Interrupt aus
	uint8_t tmp;
	tmp = ~PINB;						// lese portb ein
	tmp &= 0b00111011;					// ausmaskieren
	if ((tmp) & (KEY1)) 				// set flag
		status.key1 = 1;

	if ((tmp) & (KEY2)) 				// set flag 
		status.key2 = 1;

	if ((tmp) & (KEY3)) 				// set flag 
		status.key3 = 1;

	if ((tmp) & (KEY4)) 				// set flag
		status.key4 = 1;

	if ((tmp) & (KEY5)) 				// set flag	encoder switch
		status.key5 = 1;
} 


SIGNAL(TIMER1_COMPA_vect) 				// MAIN INTERVAL, real-time interrupt
{										// Signal(Timer_Compa_vect) Interrupt each 100 ms
	status.realtime_update = 1;			// indicate new calculation in main program
} 

SIGNAL(INT1_vect) 						// external interrupt, steigende Flanke vom Encoder Pin B 
{					
	if (PIND & (1 << PD5)) 				// welchen Wert hat nun Pin A vom Encoder? Dies bestimmt die Drehrichtung
	{
		if (last_strom > 100)
		last_strom -= 100;				// inkrementiere den Laststrom in 100 mA-Schritten
		status.encoder_ratation = 1; 	// flag for main 
	}
	else
	{
		if (last_strom < 10000)
		last_strom += 100;
		status.encoder_ratation = 1;
	}

}

	// Belegung der Ports in Controller328_Mini (gedruckte Schaltung)
	// Port B Bit 0 PB0 (PCINT0/CKL0/ICP1)		key1								
	//        Bit 1 PB1 (OC1A/PCINT1)			key2
	//        Bit 2 PB2	(SS/OC1B/PCINT3)		DI	LCD
  	//        Bit 3 PB3	(MOSI/OC2A/PCINT3))		MOSI / key3
  	//        Bit 4 PB4	(MISO/PCINT4))			MISO / key4
  	//        Bit 5 PB5 (SCK/PCINT5)			SCK  / key5 encoder
  	//        Bit 6 PB6	(PCINT6/XTAL1/TOSC1)	8MHz X-TAL
	//		  Bit 7 PB7	(PCINT7/XTAL2/TOSC2)	8MHz X-TAL

 	// Port C Bit 0 PC0 (ADC0/PCINT8)			ADC0	Batterie Controller
  	//        Bit 1 PC1	(ADC1/PCINT9)			ADC1	Voltage Shunt 0,1 Ohm
 	//        Bit 2 PC2	(ADC2/PCINT10)			ADC2	Voltage Accu DUT
  	//        Bit 3 PC3	(ADC3/PCINT11)			ADC3	nc
  	//        Bit 4 PC4	(ADC4/SDA/PCINT12)		PC4_SDA
  	//        Bit 5 PC5	(ADC5/SCL/PCINT13)		PC5_SCL
  	//        Bit 6 PC6	(PCINT/RESET)			Reset
  			          										
	// Port D Bit 0 PD0	(PCINT16/RXD)			RX			
	//        Bit 1 PD1	(PCINT17/TXD)			TX
  	//        Bit 2 PD2	(PCINT18/INT0)			Lebenslicht
  	//        Bit 3 PD3	(PCINT19/OC2B/INT1)		Encoder A
  	//        Bit 4 PD4	(PCINT20/XCK/T0)		RS LCD
  	//        Bit 5 PD5	(PCINT21/OC0B/T1)		Encoder B	
  	//        Bit 6 PD6	(PCINT22/OC0A/AIN0)		CSB	LCD		
  	//        Bit 7 PD7	(PCINT23/AIN1)			CLK	LCD			

//				PORTC 	|= (1<<PC0);			// TESTPIN ON
//				PORTC	&= ~(1<<PC0);			// TESTPIN OFF
//					dac_power_down();
//		debug();
//					uart_puts(" R_i =  ");						// UART MELDUNG
//					uart_puts((char*) ascii[3]);
//					uart_puts(" Ohm   ");						// UART MELDUNG

/*
	itoa(mA_c10th,(char*) ascii[5],10);
	uart_puts(" mA_c10th = ");			// UART MELDUNG
	uart_puts((char*) ascii[5]);

	itoa(mA_c,(char*) ascii[6],10);
	uart_puts(" mA_c = ");				// UART MELDUNG
	uart_puts((char*) ascii[6]);
*/
//	PORTC |= (1<<TESTPIN);				// TESTPIN ON
//	PORTC &= ~(1<<TESTPIN);				// TESTPIN OFF

/*
	itoa(ad_spannung_strom[1],(char*) ascii[4],10);
	uart_puts("U_0 = ");							// UART MELDUNG
	uart_puts((char*) ascii[4]);
	uart_puts("   ");

//	uart_putc(10);						// noch ein Linefeed und cr hinterher
//	uart_putc(13);
*/
/*
	itoa(ad_spannung_strom[3],(char*) ascii[4],10);
	uart_puts("U_1 = ");				// UART MELDUNG
	uart_puts((char*) ascii[4]);
	uart_putc(10);						// noch ein Linefeed und cr hinterher
	uart_putc(13);
*/
/*
void debug(void)						// output to UART for troubleshooting
{
	itoa(load_counter,(char*) ascii[4],10);
	uart_puts(" load_counter: ");		// UART MELDUNG
	uart_puts((char*) ascii[4]);

	uart_puts(" U = ");					// UART MELDUNG
	uart_puts((char*) ascii[2]);

	uart_puts(" I = ");					// UART MELDUNG
	uart_puts((char*) ascii[1]);

	uart_putc(10);						// noch ein Linefeed und cr hinterher
	uart_putc(13);
}
*/
